/*  Header computes circuit parameters for computations and proccesses on  */
/*  two-qubits.  Constructor of XVIII_Decomp produces angles for circuit of*/
/*  quant-ph/0308033 available at http://www.arXiv.org .                   */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */

#ifndef _2Q_CIRCUIT_
#define _2Q_CIRCUIT_

#include "cplx.h"
#include "oneQbit.h"
#include "twoQbit.h"
#include "2q_decomp.h"


/* Bottom conditioned controlled-not.*/
const FourByFour Csub1sup2 = 
  FourByFour (complex(1.0,0.0),complex(0.0,0.0),
	        complex(0.0,0.0),complex(0.0,0.0),
	      complex(0.0,0.0),complex(1.0,0.0),
	        complex(0.0,0.0),complex(0.0,0.0),
	      complex(0.0,0.0),complex(0.0,0.0),
	        complex(0.0,0.0),complex(1.0,0.0),
	      complex(0.0,0.0),complex(0.0,0.0),
	        complex(1.0,0.0), complex(0.0,0.0));

/* Top conditioned controlled-not. */
const FourByFour Csub2sup1 = 
 FourByFour (complex(1.0,0.0),complex(0.0,0.0),
	        complex(0.0,0.0),complex(0.0,0.0),
	      complex(0.0,0.0),complex(0.0,0.0),
	        complex(0.0,0.0),complex(1.0,0.0),
	      complex(0.0,0.0),complex(0.0,0.0),
	        complex(1.0,0.0),complex(0.0,0.0),
	      complex(0.0,0.0),complex(1.0,0.0),
	        complex(0.0,0.0), complex(0.0,0.0));

const FourByFour SWAP = Csub1sup2 * Csub2sup1 * Csub1sup2;

/* Angles for building relative phase matrix as [Rz (x) Rz] C_1^2 Rz C_1^2.*/
/* This supposes PhaseMatrix is diagonal.*/
void PhaseAngles (FourByFour Diag, double* angles)
{
  OneQubitMap Id1Q; /* one-qubit identity */
  complex z =det(Diag);
  z=sqrt(sqrt(z)); Diag = (1/z)*Diag;
  
  //angles[2]=(-0.5)*arg( 
 //		       (Diag.Matrix[1][1]*Diag.Matrix[4][4])/
	//(Diag.Matrix[2][2]*Diag.Matrix[3][3]));
 // FourByFour Tensr = (Diag * Csub1sup2) * 
      // (Tensor(Id1Q, RotZGate((-1.0)*angles[2])) * Csub1sup2);
    //angles[0]=(0.5)*(arg( (Tensr.Matrix[3][3]/Tensr.Matrix[1][1]))+
	//	   arg	(Tensr.Matrix[4][4]/Tensr.Matrix[2][2]));
  // angles[1]=(0.5)*(arg( (Tensr.Matrix[2][2]/Tensr.Matrix[1][1]))+
	//	   arg	(Tensr.Matrix[4][4]/Tensr.Matrix[3][3]));
  angles[2]=arg(Diag.Matrix[2][2])+arg(Diag.Matrix[3][3]);
  angles[0]=(-1.0)*(arg(Diag.Matrix[1][1])+arg(Diag.Matrix[2][2]));
  angles[1]=(-1.0)*(arg(Diag.Matrix[1][1])+arg(Diag.Matrix[3][3]));
};


/*  Decomposes a stochastically reversible two-qubit measurement
into C_1^2 [mu (x) 1] C_1^2 [mu (x) mu] for each mu a one-qubit
(stochastically reversible) measurement.*/
/*  Supposes Diag is real, diagonal, nonzero determinant.*/
void Meas_Param ( const FourByFour&  Diag, double* param)
{
  double a,b,c,d;
  a=Re(Diag.Matrix[1][1]); b=Re(Diag.Matrix[2][2]);
  c=Re(Diag.Matrix[3][3]); d=Re(Diag.Matrix[4][4]);
  param[0]=(b*c)/(a*d+b*c);
  param[1]=(c*d)/(a*b+c*d);
  param[2]=(b*d)/(b*d+c*a);
};

OneQubitMap oneQmeas( double sensitivity)
{
  double q = 1-sensitivity;
  return OneQubitMap ( complex(sqrt(q),0.0),
		       complex(0.0,0.0),
		       complex(0.0,0.0),
		       complex(sqrt(sensitivity), 0.0));
};


/* Structure encodes circuit for unitary U in 18 gates, Fig I, q-ph/0308033.*/
/*  Here is a sample usage:
    XVIII_Decomp Circuit = XVIII_Decomp (U);
    FourByFour Check = Circuit.global_phase* Tensor(Circuit.b,Circuit.a)*
    Csub2sup1* Tensor(OneQId, RotYGate(Circuit.right_Ry_angle)) 
    * Csub1sup2*
    Tensor(RotZGate(Circuit.top_Rz_angle),RotYGate(Circuit.left_Ry_angle))*
    Csub2sup1*Tensor(Circuit.c,Circuit.d);
*/
/* Then Check==U.*/
/*  Of course, this requires constructor be called on unitary matrix.       */
struct XVIII_Decomp
{
  complex global_phase;
  OneQubitMap a,b,c,d;
  double top_Rz_angle;
  double left_Ry_angle;
  double right_Ry_angle;

  XVIII_Decomp (void)
  {
    complex global_phase=complex (1.0,0.0);
    OneQubitMap Id;
    a=Id; b=Id; c=Id; d=Id;
    top_Rz_angle=0.0; left_Ry_angle=0.0, right_Ry_angle=0.0;
  };

  XVIII_Decomp (FourByFour U);
  /* Must be canonical decomp of SWAP v, not v. See q-ph/0303033.   */
  XVIII_Decomp (Canonical_Decomp can_decomp);
};

/* Circuit decomp for unitary, i.e. supposes U unitary.*/
/* See pg 2, III.1, q-ph/0308033. */
/* Change labels c<->a, d<->b so a(x)b leads in circuit.*/
/* u= [a(x)b] C_2^1 [R_z(top_Rz_angle) (x) R_y(left_Ry_angle)] ...
   ... C_1^2 [1 (x) R_y (right_Ry_angle)] C_2^1 [c(x)d] */
XVIII_Decomp :: XVIII_Decomp (FourByFour U)
{
  const OneQubitMap OneQId;
  complex z;
  double angles[3];
  
  FourByFour UPrime; UPrime=SWAP*U;
  /* cout << "As a check, the matrix UPrime is:" << endl << 
     UPrime << endl;*/
  
  /* Force det(UPrime)=1. */  
  z = det(UPrime); z=sqrt(sqrt(z)); UPrime=(1/z)*UPrime;
  /* cout << "As a check, the matrix UPrime is:" << endl << 
     UPrime << endl;
     cout << "The determinant is " << det(UPrime) << "." << endl;*/
  
  Canonical_Decomp DecPrime ( UPrime);

  
  /*Compute Rz angles for phasing magic basis. */
  PhaseAngles(DisEntangler*DecPrime.Phase_Magic*Entangler, /*out*/ angles);
  
  a=DecPrime.a;b=DecPrime.b;c=DecPrime.c;d=DecPrime.d;
  
  b=b*inverse(Sz); d=Sz*d;
  
  /* Checks second circuit diagram; phase may be off by +/- 1, +/- i. */
  /* FourByFour UPrime_Product = Tensor(a,b)*Csub2sup1*
     Csub1sup2 * Tensor(OneQId, RotYGate(angles[2])) * Csub1sup2
     *Tensor(RotZGate(angles[0]),RotYGate(angles[1]))*
     Csub2sup1*Tensor(c,d);*/  
  top_Rz_angle=angles[0];
  left_Ry_angle=(-1.0)*angles[1];
  right_Ry_angle=(-1.0)*angles[2];
  
  /* This block normalizes the global phase.*/
  FourByFour Product = Tensor(b,a)*
    Csub2sup1 * Tensor(OneQId, RotYGate((-1.0)*angles[2])) * Csub1sup2
    *Tensor(RotZGate(angles[0]),RotYGate((-1.0)*angles[1]))*
    Csub2sup1*Tensor(c,d);

  int count=0; int j,k; double size;
  double phase_arg=0.0; double hold;
  for(j=1;j<=4;++j)
    {
      for(k=1;k<=4;++k)
	{
	  size = complex_size(Product.Matrix[j][k]);
	  if( size > 0.01)
	    {
	      z=U.Matrix[j][k]/Product.Matrix[j][k];
	      hold=arg( U.Matrix[j][k]/Product.Matrix[j][k] ); 
	      if( fabs(hold - 2.0*PI) > 0.001)
		{
		  phase_arg+=hold; count++;
		};
	    };
	};
    };
  if(count>0)
    phase_arg=phase_arg/double(count);
  global_phase=complex(cos(phase_arg),sin(phase_arg));
};

/* Supposes canonical decomp for SWAP*u, if u is realized by output circuit.*/
XVIII_Decomp :: XVIII_Decomp (Canonical_Decomp can_dec)
{
  const OneQubitMap OneQId;
  complex z;
  double angles[3];
    
  FourByFour UPrime=Tensor(can_dec.a,can_dec.b) *
    can_dec.Phase_Magic * Tensor(can_dec.c, can_dec.d);
  PhaseAngles(DisEntangler*can_dec.Phase_Magic*Entangler, /*out*/ angles);
  a=can_dec.a;b=can_dec.b;c=can_dec.c;d=can_dec.d;
  
  b=b*inverse(Sz); d=Sz*d;
  
  top_Rz_angle=angles[0];
  left_Ry_angle=(-1.0)*angles[1];
  right_Ry_angle=(-1.0)*angles[2];
  
  /* This block normalizes the global phase.*/
  FourByFour Product = Tensor(b,a)*
    Csub2sup1 * Tensor(OneQId, RotYGate((-1.0)*angles[2])) * Csub1sup2
    *Tensor(RotZGate(angles[0]),RotYGate((-1.0)*angles[1]))*
    Csub2sup1*Tensor(c,d);

  /* Circuit realizes computation U up to phase.*/
  FourByFour U = SWAP * Tensor(can_dec.a,can_dec.b)* can_dec.Phase_Magic *
    Tensor(can_dec.c, can_dec.d);
  int count=0; int j,k; double size;
  double phase_arg=0.0; double hold;
  for(j=1;j<=4;++j)
    {
      for(k=1;k<=4;++k)
	{
	  size = complex_size(Product.Matrix[j][k]);
	  if( size > 0.01)
	    {
	      z=U.Matrix[j][k]/Product.Matrix[j][k];
	      hold=arg( U.Matrix[j][k]/Product.Matrix[j][k] ); 
	      if( fabs(hold - 2.0*PI) > 0.001)
		{
		  phase_arg+=hold; count++;
		};
	    };
	};
    };
  if(count>0)
    phase_arg=phase_arg/double(count);
  global_phase=complex(cos(phase_arg),sin(phase_arg));
};

FourByFour Eval_XVIII (XVIII_Decomp Circuit)
{
  OneQubitMap OneQId;
  return 
    (Circuit.global_phase)* 
    Tensor(Circuit.b,Circuit.a)*
    Csub2sup1* Tensor(OneQId, RotYGate(Circuit.right_Ry_angle)) 
    * Csub1sup2*
    Tensor(RotZGate(Circuit.top_Rz_angle),
	   RotYGate(Circuit.left_Ry_angle))*
    Csub2sup1*Tensor(Circuit.c,Circuit.d);
};


#endif







