/*  Header for computing two common matrix decomposition of a 4x4 complex   */
/*  matrix commonly arising in quantum computing.  First is canonical       */
/*  decomposition: any unitary v with det v =1 may be written               */
/*  as v = [b(x)c] a [d(x)f], where a applies relative phases to Bell basis.*/
/*  Header also computes singular value decomp. of any FourByFour matrix.   */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */

#ifndef _2Q_DECOMP_
#define _2Q_DECOMP_

#include "cplx.h"
#include "oneQbit.h"
#include "twoQbit.h"
#include "diag_4x4.h"


const FourByFour Entangler =  (1/sqrt(2))* 
    (FourByFour  (1,complex(0.0,1.0),0,               0,
		  0,0,               complex(0.0,1.0),complex(1.0,0.0),
		  0,0,               complex(0.0,1.0),complex(-1.0,0.0),
		  1,complex(0.0,-1.0),0,              0));

const FourByFour DisEntangler = adj(Entangler);


const OneQubitMap Sz = OneQubitMap (complex(1/sqrt(2),-1/sqrt(2)), 
				    complex(0.0,0.0),
				    complex(0.0,0.0), 
				    complex(1/sqrt(2),1/sqrt(2)));
const OneQubitMap Sx = OneQubitMap (complex(1/sqrt(2),0.0),
				    complex(0.0,(-1.0)* 1/sqrt(2)),
				    complex(0.0,(-1.0)*1/sqrt(2)),
				    complex(1/sqrt(2),0.0));

/* U in SU(4), U = [a (x) b] Phase_Magic [c (x) d]. */
/* Phase_Magic has E PhaseMagic adj(E) diagonal, E=Entangler above. */
struct Canonical_Decomp
{
  OneQubitMap a,b,c,d;
  FourByFour Phase_Magic;

  Canonical_Decomp()
  {
    OneQubitMap OneId;
    FourByFour TwoId;
    
    a=OneId; b=OneId; c=OneId; d=OneId;
    Phase_Magic=TwoId;
  };

  Canonical_Decomp (FourByFour U)
  {
    int i,j;
    complex z,w;

    FourByFour V=(DisEntangler*U)*Entangler;

    /* Force V to be norm one.*/
    w=det(V); w=sqrt(sqrt(w)); V=(1/w)*V;

    /* Matrix P^2 is complex symmetric and also unitary. */
    FourByFour PtimesP = V * transpose(V);

    /* Returned matrix O1 is real, if you trace Diagonalizer call. */
    FourByFour O1;
    O1=Diagonalizer (PtimesP);
    O1=RealGramSchmidt(O1);
    /* We want det(O1)=1.*/
    z=det(O1); O1 = z*O1;
    FourByFour D = transpose(O1) *(PtimesP *O1);
    
    for(i=1;i<=4;++i)
      D.Matrix[i][i]=sqrt(D.Matrix[i][i]);

    z=det(D)-det(V);
    if(norm(z)>.5)
      D.Matrix[1][1]=(-1.0)*D.Matrix[1][1];
    /* Det D^2=det V^2, so now det D=det V. */
    /* Whence det O3=1 below.  */

    FourByFour P = O1 * D * transpose(O1);

    FourByFour O3= adj(P)* V;
    FourByFour O2= transpose(O1)*O3;
    /* O2 must also have determinant one. */

    TensorDecomp O1Decomp (Entangler * (O1 * DisEntangler));
    a = O1Decomp.FirstFactor;
    b = O1Decomp.SecondFactor;

    TensorDecomp O2Decomp ( Entangler * (O2 * DisEntangler));
    c = O2Decomp.FirstFactor;
    d = O2Decomp.SecondFactor;

    Phase_Magic = (Entangler * (D * DisEntangler));
    /* Decomposition is now correct up to global phase.  We place
       global phase into D. */
    Phase_Magic=w*Phase_Magic;

    /* Correct so components a,b,c,d are all determinant one.*/
    z=det(a); z=sqrt(z); a=(1/z)*a; b = z*b;
    z=det(c); z=sqrt(z); c=(1/z)*c; d=z*d;
  };
};

/* Computes M = U D V, U,V unitary, D diagonal, real, >=0, for 4x4 (only) */
struct SVD_complex
{
  FourByFour LeftUni; /* Unitary U*/
  FourByFour RightUni; /*Unitary V*/
  double sing_vals[4]; /*Four Sing val's in [0]..[3] */

  SVD_complex(void)
  {
    FourByFour Id;
    int i;

    LeftUni=Id; RightUni=Id;
    for(i=0;i<=3;++i)
      sing_vals[i]=0.0;
  };

  SVD_complex (const FourByFour& W)
  {
    int i;
    FourByFour PtimesP=W*adj(W);
    /*cout << "PtimesP is " << endl << PtimesP << endl;*/
    FourByFour U = Diagonalizer(PtimesP);
    /*cout << "Diagonalizer returns U as:" << endl << U << endl;*/
    U = GramSchmidt(U);

    FourByFour check = U*adj(U);
    /*cout << "As a check, we have U*adj(U) to be " << endl << check << endl;*/
    check = adj(U)*(PtimesP *U);
    /*cout << "Also, we have adj(U) PtimesP U as " << endl << check << endl;*/
    
    FourByFour D = adj(U)*(PtimesP*U);
    for(i=1; i<=4 ;++i)
      D.Matrix[i][i]=sqrt(D.Matrix[i][i]);
    for(i=1;i<=4;++i)
      if(Re(D.Matrix[i][i])<0)
	D.Matrix[i][i]=-1.0*D.Matrix[i][i];
    
    /*cout << "The matrix D is" << endl << D << endl;*/
    
    FourByFour P=U*(D*adj(U));
    FourByFour T=inverse(P)*W;
    check=P*T;
    /*cout << "P*T is " << endl << check << endl;*/
    FourByFour V=adj(U)*T;
    
    check=V*adj(V);
    /*cout << "As a check, V * adj(V) is : " << endl << check << endl;*/

    check = U*(D*V);
    /*cout << "As a check, U*D*V is : " << endl << check << endl;*/

    LeftUni=U; RightUni=V;
    for(i=0;i<=3;++i)
      sing_vals[i]=Re(D.Matrix[i+1][i+1]);
  };
};

#endif

