/*  Header for manipulating complex number type.                           */
/*  Complex numbers may be added, multiplied, etc.  Argument      */
/*  function arg returns complex part of complex logarithm.                */

/*  Version 2.0. cout << z << endl; gives 1 digit approx.                  */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */

#ifndef _CPLX_H_
#define _CPLX_H_
#include <iostream>
#include <iomanip>
#include <math.h>

const double PI=M_PI;

using std::ostream;
using std::istream;
using std::cout;
using std::cerr;
using std::cin;
using std::endl;
using std::setprecision;
using std::setfill;
using std::setw;

//This forces parameter angles to be between 0 and 2*PI.
double ModTwoPI (double theta) { return theta-2*PI*floor(theta/(2*PI)); }

double Round (int tolerance, double input)
{
   double  TenPower=pow(10,tolerance);
   return  floor(input*TenPower+0.5) / TenPower;
}

struct complex
{
  double re, im;
  complex( double a, double b) : re(a), im(b) {}
  complex( double a ) : re(a), im(0) {}
  complex( ) : re(0), im(0) {}
};

double    Re (complex z) { return z.re ; }
double    Im (complex z) { return z.im ; }
double  norm (complex z) { return z.re*z.re+z.im*z.im; }
complex  Bar (complex z) { return complex ( z.re, - z.im); }

complex operator- (const complex& z)
{ return complex(-z.re,-z.im); }

complex operator+ (complex z, complex w)
{ return complex (z.re + w.re, z.im + w.im); }

void operator+= (complex& z, complex w)
{ z.re+=w.re; z.im+=w.im; }

complex operator- (complex z, complex w)
{ return complex (z.re - w.re, z.im - w.im); }

void operator-= (complex& z, complex w)
{ z.re-=w.re; z.im-=w.im; }

complex operator* (complex z, complex w)
{ return complex (z.re * w.re - z.im * w.im, z.im*w.re + w.im * z.re); }

void operator*= (complex& z, complex& w) { z=z*w; }

complex operator/ (complex z, complex w)
{ return complex ( (1/norm(w)) * z * Bar(w)); } 

void operator/= (complex& z, complex w) { z=z/w; }

double arg (complex z) { return ModTwoPI(atan2(z.im, z.re)); }

complex sqrt(complex z)
{
  double theta = arg(z);
  return sqrt(sqrt(z.re*z.re+z.im*z.im)) * 
    complex(cos(theta/2), sin(theta/2));
}

ostream& operator<<(ostream& out, complex z)
{
  if(z.im<0)
    return out // << setw(4) <<  setfill(' ')
	       << Round(1, (z.re)) << "-" // << setw(3) << setfill(' ')
	       << Round(1, (-1)*(z.im)) << "i";
  else
    return out // << setw(4) << setfill(' ')
	       << Round(1, (z.re)) << "+" // << setw(3) << setfill(' ')
	       << Round(1, (z.im)) << "i";
}

double complex_size( complex z)
{ return (fabs(Re(z)) + fabs(Im(z))); }
  

#endif
