/*  Functions for stepping over discrete degrees of freedom in two-qubit   */
/*  canonical decompositions.                                              */
/*  Using gate counts provided by gatecount.h, program seeks small circuits*/
/*  as the possibilities are traversed.                                    */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Standards and  Technology,                        */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */


#ifndef _DDF_
#define _DDF_

#include<iostream>
#include<sstream>
#include<string>
#include "cplx.h"
#include "oneQbit.h"
#include "twoQbit.h"
#include "2q_circuit.h"
#include "gatecount.h"

using std::string;

/* Tol for testing equality of matrix entries, output v. input.*/
const double TOL=0.01; 

/* Needed to detect when NaN is returned by floating point.*/
string convertToString(double x)
{
   std::ostringstream o;
   if (o << x)
     return o.str();
   // some sort of error handling goes here...
   return "conversion error";
};

bool Is_Bad (complex z)
{
  string re = convertToString (Re(z));
  string im = convertToString (Im(z));
  return ( //(re[0]=='N') || (im[0]=='N') 
	   //|| (re[1]=='N') || (im[1]=='N')  //Solaris 
	  (re[0]=='n') || (im[0]=='n')
	  || (re[1]=='n') || (im[1]=='n')  //Linux.
	   );
};

bool operator== (const FourByFour& A, const FourByFour& B)
{
  bool answer=true;
  for(int j=1;j<=4;++j)
    for(int k=1;k<=4;++k)
      {
	if(Is_Bad (A.Matrix[j][k]))
	   answer=false;
	if(Is_Bad (B.Matrix[j][k]))
	  answer=false;
	if(complex_size( A.Matrix[j][k]-B.Matrix[j][k]) > TOL)
	  answer=false;
      };
  return answer;
};

bool operator!= (const FourByFour& A, const FourByFour& B)
{ return !(A==B); };

const double SYNTH_TOL=10.0*DBL_EPSILON;
//const double SYNTH_TOL=0.0001;

/* Check to see if a sign flip preserves the determinant one condition.*/
bool even_flip (int signs[5])
{
  int sum=0;
  for(int j=1; j<=4; ++j)
    sum+=signs[j];
  return ( (sum==0) || (sum==2) || (sum==4));
};


/* Produces new Canonical Decomp for matrix by permuting concurrence        */
/* spectrum. Algorithm uses entanglers to shif to to unitary SVD o_1 d o_2, */
/* replacing this with (o_1 sigma^t ) sigma d sigms (sigma^t o_2).  Here,   */
/* sigma is a permutation matrix with sign flip so det (sigma)=1.           */
Canonical_Decomp Perm_can_dec ( Canonical_Decomp decomp, 
				/*in*/ int* perm, /*in*/ int*
				signs /* values 0 or 1 in 1..4 */)
{
  int j,k;
  FourByFour o1= DisEntangler * Tensor(decomp.a,decomp.b) * Entangler;
  FourByFour o2= DisEntangler * Tensor(decomp.c,decomp.d) * Entangler;
  FourByFour d= DisEntangler * decomp.Phase_Magic * Entangler;
  /* d should be diagonal unitary by construction of Phase_Magic.*/

  FourByFour temp;
  Canonical_Decomp answer;

  temp = Tensor(decomp.a,decomp.b) * decomp.Phase_Magic * Tensor(decomp.c,
								 decomp.d);
  //cout << "Check in constructor of can dec:" << endl << temp;

  temp=o1*d*o2;
  //cout << "Old o1*d*o2 on next line:" << endl << temp;

  FourByFour sigma; /* Initializes to 1; fix.*/
  for(j=1;j<=4;++j)
    sigma.Matrix[j][j]=complex(0.0,0.0);
  for(j=1;j<=4;++j)
    sigma.Matrix[j][perm[j]]=complex(1.0,0.0);
  /* For sigma to have determinant one.*/
  if(Re (det(sigma)) < 0 )
    for(j=1;j<=4;++j)
      sigma.Matrix[j][1]=(-1.0)*sigma.Matrix[j][1];
  
  if(even_flip (signs))
    for(j=1;j<=4;++j)
      if(signs[j]==1)
	for(k=1;k<=4;++k)
	  sigma.Matrix[j][k]=(-1.0)*sigma.Matrix[j][k];
  FourByFour sig_transpose=adj(sigma);
  
  o1=o1*sig_transpose; o2=sigma*o2; d=sigma*d*sig_transpose;
  
  temp=o1*d*o2;
  //cout << "New o1*d*o2 on next line:" << endl << temp;

  answer.Phase_Magic = Entangler * d * DisEntangler;
  TensorDecomp new_o1_ten_dec (Entangler * o1 * DisEntangler); 
  TensorDecomp new_o2_ten_dec (Entangler * o2 * DisEntangler);
  answer.a = new_o1_ten_dec.FirstFactor;
  answer.b = new_o1_ten_dec.SecondFactor;
  answer.c = new_o2_ten_dec.FirstFactor;
  answer.d = new_o2_ten_dec.SecondFactor;
  return answer;
};

/*  Builds 24x4 matrix (array) holding all permutation data on 4 elements.*/
/*  For conveince, based at 1, not 0.                                     */
int Build_Perms ( /* out */ int Perms[25][5] )
{
  int count=1; 
  int i,j,k,l;
  for(i=1;i<=4;++i)
    for(j=1;j<=4;++j)
      for(k=1;k<=4;++k)
	for(l=1;l<=4;++l)
	  if( (i!=j)&&(i!=k)&&(i!=l)&&(j!=k)&&(j!=l)&&(k!=l))
	    {
	      Perms[count][1]=i;Perms[count][2]=j;
	      Perms[count][3]=k;Perms[count][4]=l;
	      count++;
	    };
};

/*  Builds 8x4 matrix (array) holding all choices of signs of four elements,*/
/*  up to simulataneous quadruple sign.  Signs encoded by j, for (-1)^j.    */
/*  Based at 1 instead of 0 for convenience.                                */
int Build_Signs ( /* out*/ int Signs[9][5] )
{
  int p,q,r;  int count=1;

  for(p=0;p<=1;++p)
    for(q=0;q<=1;++q)
      for(r=0;r<=1;++r)
	{
	  Signs[count][1]=0;Signs[count][2]=p;
	  Signs[count][3]=q;Signs[count][4]=r;
	  count++;
	};
};

/* Steps over degrees of freedom in generic canonical decompositions looking*/
/* for small circuits.*/
/* Note:  Structured computations, e.g. Fourier transform on Z/4Z, may have */
/* degenerate canonical decompositions with countably infinite rather than  */
/* discrete degrees of freedom.                                             */
XVIII_Decomp can_dec_opt_circ ( const FourByFour& u)
{
  /* const; other canon's arise by modifying this basepoint decomp. */
  const Canonical_Decomp base_xu_canon (SWAP*u);
  Canonical_Decomp loop_xu_canon=base_xu_canon;
  int cur_count=42; int smallest_count = 1000;
  XVIII_Decomp cur_circ(base_xu_canon); 
  XVIII_Decomp smallest_circ =cur_circ;

  int Perms[25][5];
  Build_Perms( Perms); /* See Build_Perms comment.*/
  
  int Signs[9][5];
  Build_Signs ( Signs); /* See Build_Signs comment.*/
  FourByFour Check, Check2;
  complex z; bool agree;
  double angles[3]; OneQubitMap OneQId; 
  //OneQubitMap f; 
  //OneQubitDecomposition fred;

  for(int j=1; j<=24; ++j)
    for( int k=1;k<=8;++k)
      if(even_flip(Signs[k]))
	{
	  loop_xu_canon=Perm_can_dec (base_xu_canon, Perms[j], Signs[k]);
	  cur_circ=XVIII_Decomp (loop_xu_canon);
	  cur_circ = Clean_XVIII (cur_circ); /* Handles special cases.*/
	  cur_count= Gate_Count_XVIII (cur_circ);
	  // cout << "Gate count, loop (" << j << "," << k << ") is " <<
	  //  cur_count << "." << endl;
	  // cout << "Matrices:" << endl;
	  //cout << loop_xu_canon.a << endl;
	  //cout << loop_xu_canon.b << endl;
	  //cout << loop_xu_canon.c << endl;
	  //cout << loop_xu_canon.d << endl;
	  //cout << "angles respectively" << endl;
	  //OneQubitDecomposition a_angles ( loop_xu_canon.a );
	  //cout << a_angles << endl;
	  //OneQubitDecomposition b_angles ( loop_xu_canon.b );
	  //cout << b_angles << endl;
	  //OneQubitDecomposition c_angles ( loop_xu_canon.c );
	  //cout << c_angles << endl;
	  //OneQubitDecomposition d_angles ( loop_xu_canon.d );
	  //cout << d_angles << endl;	
	  // cout << "----------o-------------" << endl;
	  
	  Check = Eval_XVIII (cur_circ);
	  agree=(u==Check);  // constructor XVIII_Decomp adds global phase
	  if(cur_count < smallest_count)
	    if(agree) 
	      {
		smallest_count=cur_count;
		smallest_circ = cur_circ;
	      };
	  //if(!agree)
	  //{
	  //  cout << "---------------------------------------------" << endl;
	  //  cout << "ERROR; matrices don't agree!" << endl;
	  //  Check = (1/cur_circ.global_phase)*Eval_XVIII( cur_circ);
	  //   cout << "Circuit up to phase produces" << endl << Check << endl;
	  //  Check = SWAP * Tensor(loop_xu_canon.a, loop_xu_canon.b) *
	  //loop_xu_canon.Phase_Magic * 
	  //Tensor(loop_xu_canon.c,loop_xu_canon.d);
	      
	  //  cout << "Canonical decomp check." << endl << Check;
	  // Check = DisEntangler * loop_xu_canon.Phase_Magic * Entangler;
	  //   z=det(Check); z=sqrt(sqrt(z)); Check = (1/z)*Check;
	  //  cout << "Check for E^* Phase_m E" << endl << Check;
	  //  PhaseAngles (Check,/*out*/  angles);
	  //  Check2=Tensor(RotZGate(angles[0]),RotZGate(angles[1]))*
	  //Csub1sup2 * Tensor (OneQId,RotZGate(angles[2])) * Csub1sup2;
	  //  cout << "Phase angles check " << endl << Check2;
	  //  Check=Check * adj(Check2);
	  //  cout << "Check of difference in phases " << endl << Check;
	  //  cout << "--------------------------------------------" << endl;
	      //for(int l=1; l<=4;++l)
	       //{
		//for(int m=1; m<=4; ++m)
		  //{
		  //z=u.Matrix[l][m]/Check.Matrix[l][m];
		  // cout << z << " ";
		  // };
		//cout << endl;
		//};
	  // };
	  //if(agree)
	    //{
	     // cout << "Current circuit realizes u." << endl;
	      //cout << "It holds " << cur_count << " gates." << endl;
	     // Check = Tensor(loop_xu_canon.a, loop_xu_canon.b);
	     // cout << "a(x)b factor of ccd: " << endl << Check;
	      
	     // Check = Tensor(loop_xu_canon.c, loop_xu_canon.d);
	     // cout << "c(x)d factor of ccd: " << endl << Check;
	     // cout << "a factor" << endl << loop_xu_canon.a;
	     // fred=OneQubitDecomposition (loop_xu_canon.a);
	      //cout << "Decomp " << fred;
	     // cout << "b factor" << endl << loop_xu_canon.b;
	     // fred=OneQubitDecomposition (loop_xu_canon.b);
	     // cout << "Decomp " << fred;
	     // cout << "c factor" << endl << loop_xu_canon.c;
	     // fred=OneQubitDecomposition (loop_xu_canon.c);
	     // cout << "Decomp " << fred;
	     // cout << "d factor" << endl << loop_xu_canon.d;
	     // fred=OneQubitDecomposition (loop_xu_canon.d);
	     // cout << "Decomp " << fred;
	      //Check = DisEntangler * loop_xu_canon.Phase_Magic * Entangler;
	      //cout << "E^* P_M E factor:" << endl << Check;
	    //};
	};
  return smallest_circ;
};

#endif
