/*  Header for diagonalizing 4x4 complex matrices numerically, by Stephen */
/*  Bullock.  Method is to solve the characteristic polynomial numerically*/
/*  using Newton's method via polys.h, then echelon reduce v-lambda 1 to find*/
/*  the successive kernels.                                                  */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */

#ifndef _DIAG_H_
#define _DIAG_H_


#include <limits.h>
#include "cplx.h"
#include"polys.h"
#include "twoQbit.h"

const double Ech_tol=0.005;

void charpoly (FourByFour A, complex* Output)
{
//  Characteristic polynomial of a 4x4.

  Output[4]=complex(1.0,0.0);  // lambda^4 term.
  Output[3]=(-1.0)*(A.Matrix[1][1]+A.Matrix[2][2]+A.Matrix[3][3]
		    +A.Matrix[4][4]);  //lambda^3 term.
  Output[2]=(A.Matrix[1][1] *A.Matrix[2][2]- A.Matrix[3][1]*A.Matrix[1][3]
	     - A.Matrix[2][1]*A.Matrix[1][2] - A.Matrix[4][1]*A.Matrix[1][4]
	     + A.Matrix[1][1]*A.Matrix[3][3] + A.Matrix[2][2]*A.Matrix[3][3]
	     + A.Matrix[2][2]*A.Matrix[4][4] + A.Matrix[3][3]*A.Matrix[4][4]
	     - A.Matrix[3][4]*A.Matrix[4][3] - A.Matrix[3][2]*A.Matrix[2][3]
 	     - A.Matrix[4][2]*A.Matrix[2][4] + A.Matrix[1][1]*A.Matrix[4][4]);
  //That was the lambda^2 term.
  Output[1]=(A.Matrix[1][1]*A.Matrix[3][2]*A.Matrix[2][3]
	     - A.Matrix[3][2]*A.Matrix[2][4]*A.Matrix[4][3]
	     - A.Matrix[2][1]*A.Matrix[4][2]*A.Matrix[1][4]
	     - A.Matrix[1][1]*A.Matrix[3][3]*A.Matrix[4][4]
	     - A.Matrix[2][1]*A.Matrix[3][2]*A.Matrix[1][3]
	     - A.Matrix[4][2]*A.Matrix[2][3]*A.Matrix[3][4]
	     - A.Matrix[2][2]*A.Matrix[3][3]*A.Matrix[4][4]
	     - A.Matrix[1][1]*A.Matrix[2][2]*A.Matrix[3][3]
	     + A.Matrix[2][1]*A.Matrix[1][2]*A.Matrix[4][4]
	     - A.Matrix[1][1]*A.Matrix[2][2]*A.Matrix[4][4]
	     + A.Matrix[4][2]*A.Matrix[2][4]*A.Matrix[3][3]
	     + A.Matrix[3][1]*A.Matrix[2][2]*A.Matrix[1][3]
	     - A.Matrix[3][1]*A.Matrix[1][2]*A.Matrix[2][3]
	     + A.Matrix[3][1]*A.Matrix[1][3]*A.Matrix[4][4]
	     - A.Matrix[3][1]*A.Matrix[1][4]*A.Matrix[4][3]
	     - A.Matrix[4][1]*A.Matrix[1][2]*A.Matrix[2][4]
	     + A.Matrix[2][2]*A.Matrix[3][4]*A.Matrix[4][3]
	     + A.Matrix[3][2]*A.Matrix[2][3]*A.Matrix[4][4]
	     + A.Matrix[1][1]*A.Matrix[4][2]*A.Matrix[2][4]
	     - A.Matrix[4][1]*A.Matrix[1][3]*A.Matrix[3][4]
	     + A.Matrix[4][1]*A.Matrix[1][4]*A.Matrix[3][3]
	     + A.Matrix[1][1]*A.Matrix[3][4]*A.Matrix[4][3]
	     + A.Matrix[4][1]*A.Matrix[2][2]*A.Matrix[1][4]
	     + A.Matrix[2][1]*A.Matrix[1][2]*A.Matrix[3][3]); //lambda^1

  Output[0]=det(A);
}

/* Store column space of dimension 0,1,2,3 by cols in FourByFour.*/
struct ColSpace
{
  FourByFour Cols;
  int dim;
};


FourByFour Echelon (FourByFour V)
{
 
  int i,j,k,l,m;
  int s,t;
  complex z;

  int ColumnNumber=0,RowNumber=0;  //Row & Col of last pivot entry.

  double EntrySize; double Largest;

  for(j=1;j<=4;++j)
    {
      EntrySize=0.0;
      Largest=0.0;
      m=0;
      for(i=RowNumber+1;i<=4;++i)
	{
	  EntrySize=fabs(Re(V.Matrix[i][j]))+fabs(Im(V.Matrix[i][j]));
	  if(EntrySize>Largest)
	    {m=i; Largest=EntrySize;}
	} // Matrix[m][j] now holds largest eligible entry of col j
      if(Largest < Ech_tol)  //Clear column j if no pivot.
	{
	  for(i=RowNumber+1;i<=4;++i)
	    V.Matrix[i][j]=complex(0.0,0.0);
	  for(i=1;i<=RowNumber;++i)
	    {	      
	      if(fabs(Re(V.Matrix[i][j])) 
		 + fabs(Im(V.Matrix[i][j])) < Ech_tol)
		V.Matrix[i][j]=complex(0.0,0.0);
	    }
	}
      if(Largest >= Ech_tol)  //Matrix[i][j] is declared a pivot.
	{ //flip row of pivot to highest left possible so as to not interfere.
	  RowNumber++; ColumnNumber=j;
	  z=V.Matrix[m][ColumnNumber];
	  for(k=1;k<=4;++k)
	    {
	      z=V.Matrix[RowNumber][k];
	      V.Matrix[RowNumber][k]=V.Matrix[m][k];
	      V.Matrix[m][k]=z;
	    }
	  //Clear row by dividing out pivot.
	  z=V.Matrix[RowNumber][ColumnNumber];
	  for(l=1;l<=4;++l)
	    V.Matrix[RowNumber][l]=V.Matrix[RowNumber][l]/z;

	  for(k=1;k<=4;++k)  //Use pivot to clear its column.
	    {
	      if(k!=RowNumber)
	      {
		z=V.Matrix[k][ColumnNumber];
		for(l=1;l<=4;++l)
		  V.Matrix[k][l]=V.Matrix[k][l]-z*V.Matrix[RowNumber][l];
	      }
	    }
	}
    }
  return V;
};

/* Kernel returns kernel of input as struct ColSpace.*/
/* WARNING:  Supposes A is Echelon reduced!  Call Echelon first.*/
ColSpace Kernel (const FourByFour& A)
{
  FourByFour Reduced=A; Reduced=Echelon(A); /* A passed as constant.*/
  ColSpace Answer;
  Answer.dim=1;

  bool Is_Pivot[5][5]; /* Set to true if position is pivot of Reduced.*/
  bool col_has_no_pivot[5];
  bool col_is_zero[5];

  int j,k,l;

  /* Intialize matrix in answer to be 0.*/
  for(j=1;j<=4;++j)
    for(k=1;k<=4;++k)
      Answer.Cols.Matrix[j][k]=complex(0.0,0.0);

  for(j=1;j<=4;++j)
    for(k=0;k<=4;++k)
      Is_Pivot[j][k]=false;

  /* Move left to right on each row.  First nonzero entry is pivot.*/
  k=0;
  for(j=1;j<=4;++j)
    while( (!Is_Pivot[j][k]) && k<=3)
    {
      k++;
      /* cout << "complex_size(Reduced.Matrix[" << j << "][" << k << "] is "
	<< complex_size(Reduced.Matrix[j][k]) << "." << endl; */
      if( complex_size(Reduced.Matrix[j][k]) > Ech_tol) 
	Is_Pivot[j][k]=true;
    };

  /* cout << "Is_Pivot is:" << endl;
     for(j=1;j<=4;++j)
     {
     for(k=1;k<=4;++k)
     cout << Is_Pivot[j][k];
     cout << endl;
     };*/

  for(k=1;k<=4;++k)
    { 
      col_has_no_pivot[k]=true;     
      for(j=1;j<=4;++j)
	if(Is_Pivot[j][k])
	  col_has_no_pivot[k]=false;
    };

  for(k=1;k<=4;++k)
    { 
      col_is_zero[k]=true;     
      for(j=1;j<=4;++j)
	if(complex_size(Reduced.Matrix[j][k]) > Ech_tol )
	  col_is_zero[k]=false;
    };

  /* cout << "col_is_zero is ";
     for(k=1;k<=4;++k)
  cout << col_is_zero[k] << " ";
  cout << endl;*/
  
  /*cout << "col_has_no_pivot is ";
    for(k=1;k<=4;++k)
    cout << col_has_no_pivot[k] << " ";
    cout << endl;*/
  
  for(k=1;k<=4;++k)
    if(col_is_zero[k])
      {
	Answer.Cols.Matrix[k][Answer.dim]=complex(1.0,0.0);
	Answer.dim++;
	/*cout << "Writing basis vector, 1 at (" << k << "," <<
	  (Answer.dim-1) << ")." << endl;*/
      };

  for(k=1;k<=4;++k)
    if( (!col_is_zero[k]) && (col_has_no_pivot[k]) )
      {
	for(j=1;j<=4;++j)
	  if(complex_size(Reduced.Matrix[j][k])>Ech_tol )
	    {
	      /* Find l so that (l,j) is a pivot.*/
	      l=1;
	      while (! Is_Pivot[j][l])
		l++;
	      /* Write nonzero entry of col k.*/
	      Answer.Cols.Matrix[l][Answer.dim]=Reduced.Matrix[j][k];
	      Answer.Cols.Matrix[k][Answer.dim]=(-1.0);
	      /*cout << "Printing two term entry, 1 at (" <<
		l << "," << Answer.dim << "), and "
		<< Reduced.Matrix[j][k] << " at (" << j << "," << Answer.dim
		<< ")." << endl;*/
	    }
	Answer.dim++;
      };
       
  Answer.dim--;
  return Answer;
};

void CleanList (complex* Input, bool* Changes)
  // Changes[i] is true if, after cleaning, Input[i]!=Input[i+1].
  // Code supposes Input[1..4].
{
  complex Copy[4];
  complex ExtraCopy[4];
  int Population[4];
  bool SameAs[4][4];
  int i,j,k,l;
  int MaxPop=1;
  complex z,w;

  for(i=0;i<=3;++i)
    Copy[i]=Input[i];

  for(i=0;i<=3;++i)
    for(j=0;j<=3;++j)
      {
	z=Copy[i]-Copy[j];
	if(fabs(Re(z))+fabs(Im(z)) < Ech_tol)
	  SameAs[i][j]=true;
	else
	  SameAs[i][j]=false;
      }
  //  for(i=1;i<=4;++i)
  //{
  //  for(j=1;j<=4;++j)
  //cout << SameAs[i][j] << " ";
  //  cout << endl;
  //}

  for(i=0;i<=3;++i)
    {
      z=Copy[i];
      Population[i]=1;
      for(j=0;j<=3;++j)
	{
	  if(SameAs[i][j])
	    {
	      if(i!=j)
		{
		  (Population[i])++;
		  z=z+Copy[j];
		  if(Population[i]>MaxPop)
		    MaxPop=Population[i];
		}
	    }
	}
      if(Population[i]>1)
	{
	  z=z/double(Population[i]);
	  for(j=i;j<=3;++j)
	    if(SameAs[i][j])
	      Copy[j]=z;
	}
    }
  //  for(i=1;i<=4;++i)
  //cout << "The value of Population[" << i <<"] is " << Population[i] << "."
  // << endl;
  k=1; l=1;
  for(i=0;i<=3;++i)
    {
      if(Population[i]>k)
	{
	  k=Population[i];
	  l=i;
	}
    }
  // cout << "The Max Pop is " << MaxPop << "." << endl;
      
  if(MaxPop>1)
   {
     j=0;
     for(k=0;k<=3;++k)
       {
	 if(SameAs[k][l])
	   {
	     ExtraCopy[j]=Copy[k];
	     j++;
	   }
       }
     for(k=0;k<=3;++k)
       {
	 if(!SameAs[k][l])
	   {
	     ExtraCopy[j]=Copy[k];
	     j++;
	   }
       }
     for(i=0;i<=3;++i)
       Input[i]=ExtraCopy[i];
   }    
  for(i=0;i<=2;++i)
    {
      z=Input[i]-Input[i+1];
      if(fabs(Re(z))+fabs(Im(z)) > Ech_tol)
	Changes[i]=true;
      else
	Changes[i]=false;
    }
}


//If input diagonalizable V, returns S with S^{-1} V S diagonal.
//Algorithm is a bit of a mess.  Idea is to build S column by column
//using successive echelon reductions of V-lambda I.
//Lots of degenerate cases to check for repeated eigenvalues etc.

FourByFour Diagonalizer (const FourByFour& V)
{
  FourByFour EVectorAsCols;
  FourByFour Reduced;
  ColSpace Ker_A_less_lambda;
  int Insert_Col=1;
  bool error=false;

  const FourByFour Id; /* Initializes to 1.*/

  int CurrentRoot=0;

  complex CharVPoly[5];
  complex Roots[4];
  bool FlipRoots[2];

  int i,j,k,l,m;

  charpoly (V, CharVPoly);
  /* cout << "Attempting to solve quartic." << endl; */
  RootsOfQuartic(CharVPoly, Roots);
  /*cout << "Solved quartic." << endl;*/
  /*for(i=0;i<=3;++i)
    cout << "Evalue #" << i << " is " << Roots[i] << " with argument " 
	 << arg(Roots[i]) << "." << endl;*/
  
  CleanList( Roots, FlipRoots);

 /*cout << "Solved quartic." << endl;*/
  /*for(i=0;i<=3;++i)
     cout << "After cleanroots, Evalue #" << 
       i << " is " << Roots[i] << " with argument " 
	  << arg(Roots[i]) << "." << endl;*/

  while(CurrentRoot<=3)
    {
      Reduced=Echelon( (V +(-1.0)*(Roots[CurrentRoot])*Id));
      Ker_A_less_lambda = Kernel(Reduced); /* Roots[CR] eigenspace*/
      
      for(k=1;k<=Ker_A_less_lambda.dim;++k)
	{
	  for(j=1;j<=4;++j)
	    EVectorAsCols.Matrix[j][Insert_Col]
	      =Ker_A_less_lambda.Cols.Matrix[j][k];
	  Insert_Col++;
	};      
      /* Block advances to next distinct root, given CleanRoots above.*/
      while(!FlipRoots[CurrentRoot])
	CurrentRoot++;
      CurrentRoot++;
    };

  //cout << "Value of Insert_Col is " << Insert_Col << "." << endl;
  //Insert_Col--;
  //cout << "Value of Insert_Col is " << Insert_Col << "." << endl;  
  //error = !(Insert_Col==4);
  

  if(error==true)
    {
      FourByFour Temp=V;
      cout << "WARNING!!! ERROR IN DIAGONALIZER ROUTINE!!!" << endl;
      //<< "Attempt to diagonalize following matrix failed:" << endl
	  // << Temp << endl;
    };
  return EVectorAsCols;
};


//This replaces all entries by real part and then uses Gram-Schmidt
//to produce matrix in O(4).  Then flips first col so output is in SO(4),
//if needed.

FourByFour RealGramSchmidt (FourByFour V)
{
  complex w;
  int i,j,k,l;
  double InnerColJK; double SquareNormColJ;

  for(i=1;i<=4;++i)
    for(j=1;j<=4;++j)
      V.Matrix[i][j]=complex(Re(V.Matrix[i][j]),0.0);
  
  //Now make each column orthogonal to others.
  for(j=1;j<=4;++j)
    for(k=j+1;k<=4;++k)
      {
	InnerColJK=0.0; SquareNormColJ=0.0;
	for(i=1;i<=4;++i)
	  {
	    InnerColJK+= Re(V.Matrix[i][j])*Re(V.Matrix[i][k]);
	    SquareNormColJ+= Re(V.Matrix[i][j])*Re(V.Matrix[i][j]);
	  }
	if(InnerColJK>Ech_tol)
	  for(i=1;i<=4;++i)
	    V.Matrix[i][k]=V.Matrix[i][k]
	      -(InnerColJK/SquareNormColJ)*V.Matrix[i][j];
      }
  for(j=1;j<=4;++j)
    {
      SquareNormColJ=0.0;
      for(i=1;i<=4;++i)
	{ SquareNormColJ+= Re(V.Matrix[i][j])*Re(V.Matrix[i][j]);}
      for(i=1;i<=4;++i)
	V.Matrix[i][j]=(1/sqrt(SquareNormColJ))*V.Matrix[i][j];
    }
  w=det(V);
  if(Re(w)<0.0)
    for(i=1;i<=4;++i)
      V.Matrix[i][1]=(-1.0)*V.Matrix[i][1];

  return V;
}

//  Applies GramSchmidt so the cols of input become Hermitian orthonormal.

FourByFour GramSchmidt (FourByFour V)
{
  complex w;
  int i,j,k,l;
  complex InnerColJK; double SquareNormColJ;
  
  //Now make each column orthogonal to others.
  for(j=1;j<=4;++j)
    for(k=j+1;k<=4;++k)
      {
	InnerColJK=complex(0.0,0.0); SquareNormColJ=0.0;
	for(i=1;i<=4;++i)
	  {
	    InnerColJK+= Bar(V.Matrix[i][j])*(V.Matrix[i][k]);
	    SquareNormColJ+= norm(V.Matrix[i][j]);
	  }
	if(norm(InnerColJK)>0.02)
	  for(i=1;i<=4;++i)
	    V.Matrix[i][k]=V.Matrix[i][k]
	      -(InnerColJK/SquareNormColJ)*V.Matrix[i][j];
      }
  for(j=1;j<=4;++j)
    {
      SquareNormColJ=0.0;
      for(i=1;i<=4;++i)
	{ SquareNormColJ+= norm(V.Matrix[i][j]);}
      for(i=1;i<=4;++i)
	V.Matrix[i][j]=(1/sqrt(SquareNormColJ))*V.Matrix[i][j];
    }
  return V;
}

#endif






