/*  This header computes gate counts for two-qubit circuit types of 2q.h     */
/*  For a given type ???_Decomp, Clean_????_Decomp  handles special cases.   */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */

#ifndef _GATECOUNT_
#define _GATECOUNT_

#include "cplx.h"
#include "oneQbit.h"
#include "twoQbit.h"
#include "2q_circuit.h"

const double TRIV_TOL=0.001;

#ifndef _DIAG_H_
double complex_size( complex z)
{ return (fabs(Re(z)) + fabs(Im(z))); }
#endif

/*  Checks whether a one-qubit-map is trivial or not. */
bool Is_Trivial_1q (const OneQubitMap& v)
{
  bool answer = false;
  complex z = det(v);
  z = sqrt(z);
  OneQubitMap u=(1/z)*v;  /* u is now determinant one.*/

  if( complex_size (v.ab - complex(0.0,0.0)) < TRIV_TOL)
    if( complex_size (v.ba - complex(0.0,0.0)) < TRIV_TOL)
      {
	u=u*u;  /* -1 is also trivial. */
	if( complex_size (v.aa - complex(1.0,0.0)) < TRIV_TOL)
	  if( complex_size (v.bb - complex(1.0,0.0)) < TRIV_TOL)
	    answer=true;
      };
  return answer;
};

/* This returns true iff theta if a multiple of 2Pi up to tolerance.        */
/* In that case, feeding theta to R_y or R_z returns a global phase comp.   */
bool ignore_angle ( double theta)
{
  double temp = ModTwoPI (theta);
  return ( (fabs(temp - 0.0) < TRIV_TOL) || (fabs(temp - (2.0)*PI)< TRIV_TOL));
};

bool count_angle ( double theta)
{  return ! ignore_angle (theta); }

/*  Scrubs 18 gate circuits for unitaries in degenerate cases.          */
XVIII_Decomp Clean_XVIII (XVIII_Decomp input_circ) /*Do NOT make const& .*/
{
  bool no_angles;
  no_angles = ignore_angle( input_circ.top_Rz_angle) &&
    ignore_angle(input_circ.left_Ry_angle) && 
    ignore_angle( input_circ.right_Ry_angle);  /* Then center block is swap.*/
  if(no_angles)
    {
      if( fabs(input_circ.top_Rz_angle - 2.0*PI) < TRIV_TOL)
	{
	  input_circ.global_phase = complex(-1.0,0.0)*input_circ.global_phase;
	  input_circ.top_Rz_angle = 0.0;
	};
      if( fabs(input_circ.left_Ry_angle - 2.0*PI) < TRIV_TOL)
	{
	  input_circ.global_phase = complex(-1.0,0.0)*input_circ.global_phase;
	  input_circ.left_Ry_angle = 0.0;
	};
      if( fabs(input_circ.right_Ry_angle - 2.0*PI) < TRIV_TOL)
	{
	  input_circ.global_phase = complex(-1.0,0.0)*input_circ.global_phase;
	  input_circ.right_Ry_angle = 0.0;
	};
      OneQubitMap OneQId;  /* Next lines push sides through SWAP.*/
      input_circ.b=input_circ.b*input_circ.d;
      input_circ.a=input_circ.a*input_circ.c;
      input_circ.c=OneQId;  input_circ.d=OneQId;
    };
  return input_circ;
};


/* OneQubitDecomposition already handles degenerate Rz Ry Rz cases with     */
/* Ry = -i \sigam^y.                                                        */
int Gate_Count_1q ( const OneQubitMap& v)
{
  int count=0;
  
  OneQubitDecomposition v_angles ( v);
  
  if ( count_angle(v_angles.alpha))   
    count++;
  if ( count_angle(v_angles.beta))
    count++;
  if (count_angle (v_angles.theta))
    count++;
  //cout << "alpha=" << v_angles.alpha << ", beta=" << v_angles.beta <<
   // ", theta=" << v_angles.theta << endl;
  return count;  
};

/* Count gates in the eighteen gate decomposition.  */
int Gate_Count_XVIII (XVIII_Decomp Circ18)
{
  int count=3; /* By default count all CNOT's.*/

  /* Not used.  These bools would check if a swap can be pushed through.  */
  //bool block2 = 
  // ignore_angle( Circ18.top_Rz_angle) && ignore_angle( Circ18.left_Ry_angle);
  //bool block3 = ignore_angle( Circ18.right_Ry_angle);

  count+= Gate_Count_1q (Circ18.a); count+= Gate_Count_1q (Circ18.b);
  count+= Gate_Count_1q (Circ18.c); count+= Gate_Count_1q (Circ18.d);

  if (count_angle (Circ18.top_Rz_angle))
    count++;
  if (count_angle (Circ18.left_Ry_angle))
    count++;
  if (count_angle (Circ18.right_Ry_angle))
    count++;
  return count;
};

#endif





