/*  Header for manipulation of 2x2 unitary matrices, interpreted as maps on*/
/*  one quantum bit.                                                       */
/*  In addition to composing maps with operator * (overloaded,) header may */
/*  also factor a determinant one computation as a product of ZYZ Bloch    */
/*  sphere rotations.                                                      */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */


#ifndef _ONEQBIT_H_
#define _ONEQBIT_H_

#include "cplx.h"

const double OneQ_TOL=(0.000001);

struct OneQubitMap
{
  complex aa, ab, ba, bb;

  OneQubitMap (complex a, complex b, complex c, complex d):
	  aa(a),ab(b),ba(c),bb(d) {}
  OneQubitMap(): aa(1),ab(0),ba(0),bb(1) {}
};

OneQubitMap transpose(const OneQubitMap& M)
{ return OneQubitMap (M.aa, M.ba, M.ab, M.bb); }

OneQubitMap operator*(const OneQubitMap& U, const OneQubitMap& V)
{
  return OneQubitMap( U.aa * V.aa + U.ab * V.ba, U.aa * V.ab + U.ab * V.bb,
		      U.ba * V.aa + U.bb * V.ba, U.ba * V.ab + U.bb * V.bb);
}
OneQubitMap operator* (complex z, const OneQubitMap& U)
{ return OneQubitMap (z*U.aa, z*U.ab, z*U.ba, z*U.bb); }

complex det (const OneQubitMap& A)
{ return (A.aa * A.bb - A.ba * A.ab);}

OneQubitMap RotZGate (double alpha)
{
  return OneQubitMap (complex (cos (alpha/2), (-1.0)*sin(alpha/2)), 
		      complex(0.0,0.0),
		      complex (0.0,0.0), 
		      complex (cos (alpha/2), sin(alpha/2)));
}

OneQubitMap RotYGate (double theta)
{
  return OneQubitMap (complex (cos(theta/2),0.0), complex (sin(theta/2),0.0),
		      complex ( -1.0*sin(theta/2),0.0), 
		      complex (cos(theta/2),0.0));
}

OneQubitMap PhaseGate (double delta)
{
  return OneQubitMap (complex (cos(delta),sin(delta)), complex (0.0,0.0),
		      complex (0.0,0.0), complex (cos(delta),sin(delta)));
}

complex TopLeft(OneQubitMap A)  { return A.aa; }
complex TopRight(OneQubitMap A) { return A.ab; }
complex BotLeft(OneQubitMap A)  { return A.ba; }
complex BotRight(OneQubitMap A) { return A.bb; }

ostream& operator<<(ostream& out, const OneQubitMap& A)
{ return out << "( " << Round(6,Re(TopLeft(A))) << "+" << 
    Round(6,Im(TopLeft(A))) << "i"
	     << " & " 
	     << Round(6,Re(TopRight(A))) << "+" << 
    Round(6,Im(TopRight(A))) << "i"  
	     << " \\" << "\\ " << endl
	     << Round(6,Re(BotLeft(A))) << "+" << 
    Round(6,Im(BotLeft(A))) << "i"  
	     << " & " 
	     << Round(6,Re(BotRight(A))) << "+" << 
    Round(6,Im(BotRight(A))) << "i"  
	     << " \\" << "\\ )" 
  << endl; }

OneQubitMap inverse(const OneQubitMap& A)
{
  complex w = det(A);
  if (norm(w)<1e-10)
  {
      cerr << "Trying to compute inverse of a degenerate matrix\n" <<endl;
      exit(1);
  }
  return OneQubitMap(A.bb/w, -A.ab/w, -A.ba/w, A.aa/w);
}

struct OneQubitDecomposition  // formerly "Angles"
{
  double delta, alpha, beta, theta;
  
  OneQubitDecomposition(const OneQubitMap& A);
  OneQubitDecomposition()
  { delta=0.0; alpha=0.0; theta=0.0; beta=0.0; };
};

ostream& operator<<(ostream& out, const OneQubitDecomposition& oqd)
{
  return out << " delta=" << oqd.delta <<endl
       << " alpha=" << oqd.alpha <<endl
       << "  theta=" << oqd.theta  <<endl
       << " beta=" << oqd.beta <<endl;
}

//This forces parameter angles to be between 0 and 4*PI.
double ModFourPI (double theta) { return theta-4*PI*floor(theta/(4*PI)); }

/*  Angles per pg.31 of "Reducing Quantum Computations to Elementary Unitary*/
/*  Operations," George Cybenko, \emph{Computers in Science and Engineering"*/
/*  March/April 2001, starting pg. 27.                                      */

OneQubitDecomposition::OneQubitDecomposition(const OneQubitMap& A)
{
  const OneQubitMap X(0.0,1.0,1.0,0.0);
  complex z, w=det(A);
  delta = arg(w)/2;
  OneQubitMap C,D, B = PhaseGate(-delta)*A;
  C = transpose(B) * X * B;
  beta = arg(BotRight(C));
  D = B * RotZGate(-beta); 
  if (complex_size(BotRight(D))<OneQ_TOL)
    alpha = (2.0) * arg(-1.0 * BotLeft(D));
  else
    if(complex_size(BotLeft(D))<OneQ_TOL)
      alpha = (2.0) * arg(BotRight(D));
    else
      alpha = arg((-1.0) * BotLeft(D)) + arg(BotRight(D));
  z = complex (cos(alpha/2),sin(alpha/2));
  w = complex( Re(TopLeft(D)*z), Re(TopRight(D)*z));
  theta = 2.0*arg(w);

  //Given the parameterization, mod4PI is allowed but not mod2PI.
  alpha=ModFourPI(alpha);
  beta=ModFourPI(beta);
  theta=ModFourPI(theta);

  //This is an occassionally useful optimization.  The net effect is to
  //flip off diagonals of the RotYGate, so RotZs are superfluous.
  if( (alpha-3*PI)*(alpha-3*PI)<0.0001)
    if( complex_size (beta-PI) < OneQ_TOL)
      {
	alpha = 0; beta = 0; theta=(-1.0)*theta; ModFourPI(theta);
      }
};

#endif
