/*  Header for sovling quartics numerically, by Stephen Bullock.           */
/*  This header's functions are intended to solve fourth degree polynomials*/
/*  using Newton's method. The original application is solving             */
/*  characteristic polynomials of 4x4 matrices before diagonalization.     */

/*  WARNING:  RANDOM NUMBER CALLS MUST BE FUNCTIONING FOR THIS HEADER TO   */
/*  RELIABLY PRODUCE CORRECT ANSWERS.  Test the code at line 64 before use */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */

#ifndef _POLYS_H_
#define _POLYS_H_

#include <float.h>
#include <stdlib.h>
#include <limits.h>
#include "cplx.h"


const double NEWTONtol=(2.0)*DBL_EPSILON; //How good f~0 before N.method quits?

/*These are needed by the Newtons method routine.*/
int rand(void);
const unsigned meshSize=RAND_MAX;
double rand0t1()           { return (rand()/double(meshSize)); };

complex SubPoly (complex z, int degree, complex* Coefficients)
{
  complex w(1.0);
  complex answer=Coefficients[0];

  for(int i=1; i<=degree; ++i)
  {  
      w*=z;
      answer+= (w*Coefficients[i]);
  }
  return answer;
}
    
void PrintPoly (int degree, complex* Coefficients)
{
  int i;
  cout << "(" << Coefficients[0] << ")+";
  for(i=1; i<=degree-1; ++i)
    cout << "(" << Coefficients[i] << ")*z^" << i << "+";
  cout << "(" << Coefficients[degree] << ")*z^" << degree;
  cout << endl;
}

void DerPoly (int degree, complex* InputPoly, complex* Derivative)
  /*degree is the degree of the InputPoly.  Derivative[degree] is not used.*/
{  for(int i=0;i<=degree-1;++i) Derivative[i]=complex(i+1)*InputPoly[i+1]; }


complex NewtonFindPolyRoot (int degree, complex* Coefficients)
{
  complex z,w;
  double a,b;
  int i,j;
  complex DerivativeCoefficients[degree+3];
  complex Copy[degree+3];
  double error=15.0;
  int count=0;

  if(degree==1)
    {
      //cout << 
      //"Returning SPECIAL " << ((-1.0)*Coefficients[0]/Coefficients[1])
      //   << endl;
      return ((-1.0)*Coefficients[0]/Coefficients[1]);
    }

  for(j=0;j<=degree;++j) Copy[j]=Coefficients[j];
  for(j=0;j<=degree;++j) Copy[j]=Copy[j]/Copy[degree];

  /* cout << "Calling FindPolyRoot on this poly:" << endl;
  PrintPoly (degree, Copy); */
  DerPoly( degree, Copy, DerivativeCoefficients);
  /* cout << "The derivative polynomial is " << endl; */
  /* PrintPoly (degree-1, DerivativeCoefficients); */

  a=rand0t1(); b=rand0t1();
  w=complex (a,b);
  z=(Copy[0]+rand0t1())*w;
  /* cout << "Seed z for first plug in is " << z << "." << endl; */

  while( fabs(error) > NEWTONtol)
    {
      if(fabs(error) > 1e6)
	{
	  a=rand0t1(); b=rand0t1();
	  w=complex (a,b);
	  z=(Copy[0]+rand0t1())*w;
	}
      /* cout << "The value of w is : " << w << "." << endl; */
      w=(SubPoly(z,degree-1,DerivativeCoefficients));
      z = z-SubPoly(z,degree,Copy)/w;
      if(norm(w)<(0.1)*NEWTONtol)
	{
	  /*cout << "CHOOSING NEW RANDOM z"; */
	  a=rand0t1(); b=rand0t1();
	  w=complex (a,b);
	  z=(Copy[0]+rand0t1())*w;
	  /* cout << " of " << z <<"." << endl; */
	}
      /*cout << "New z to plug is " << z << "." << endl; */
      w = SubPoly(z,degree,Copy);
      error = fabs(Re(w))+fabs(Im(w));
      /* cout << "NewtonIt error " << error << endl; */
      count++;
      if(count>1000)
	break;
    }  
  
  /*cout << "Returning the following root: " << z << "." << endl << endl;*/
  return z;
}

//Divides by z-Root.  degree is larger #.
void ComputeQuotientPoly( complex Root, int degree,
		   complex* InputPoly, complex* QuotientPoly)
{
  
  QuotientPoly[degree-1]=InputPoly[degree];
  for(int i=degree-1;i>=0;--i)
    { QuotientPoly[i]=InputPoly[i+1]+(Root*QuotientPoly[i+1]); }
}

void RootsOfQuartic(complex* Coefficients, complex* Roots)
{
  complex FirstPoly[5];
  complex SecondPoly[5];
  complex ThirdPoly[5];
  complex z,w;

  /*PrintPoly(4,Coefficients);*/
  w=NewtonFindPolyRoot (4, Coefficients);
  Roots[0]=w;
  ComputeQuotientPoly (w,4,Coefficients, FirstPoly);
  //PrintPoly(3, FirstPoly);
  w=NewtonFindPolyRoot (3,FirstPoly);
  //cout << "The next root is " << w << "." << endl;
  Roots[1]=w;
  ComputeQuotientPoly (w,3,FirstPoly,SecondPoly);
  //PrintPoly(2,SecondPoly);
  w=NewtonFindPolyRoot(2,SecondPoly);
  //cout << "The next root is " << w << "." << endl;
  Roots[2]=w;
  ComputeQuotientPoly (w,2,SecondPoly,ThirdPoly);
  w=NewtonFindPolyRoot(1,ThirdPoly);
  //cout << "The last root is " << w << "." << endl;
  Roots[3]=w;

  return;      
}

void SortRootsByComplexArgument (complex* roots)
{
  complex w;
  for(int i=1;i<=3;++i) //i counts passes of bubble sort.
    for(int j=0;j<=2;++j) //j counts roots.
      if(arg(roots[j])>arg(roots[j+1]))
	{ w=roots[j]; roots[j]=roots[j+1]; roots[j+1]=w; }
}

#endif

