/*  Header for manipulating four by four complex matrices, usually 4x4      */
/*  unitaries interpreted as quantum computations on two bits.              */
/*  Operations *,+ etc. also valid for any FourByFour.                      */
/*  Inverse function produces inverse in all cases, computed by echelon     */
/*  reduction.  Function adj(-) is faster and computes                      */
/*  complex-conjugate-transpose.                                            */

/*  Version 2.0.                                                            */

/* Mathematical and Computational Sciences Division                        */
/* National Institute of Technology,                                       */
/* Gaithersburg, MD USA                                                    */

/* This software was developed at the National Institute of Standards and  */
/* Technology (NIST) by employees of the Federal Government in the course  */
/* of their official duties. Pursuant to title 17 Section 105 of the       */
/* United States Code, this software is not subject to copyright protection */
/* and is in the public domain. NIST assumes no responsibility whatsoever   */
/* for its use by other parties, and makes no guarantees, expressed or     */
/* implied about its quality, reliability, or any other characteristic.    */

#ifndef _TWOQBIT_H_
#define _TWOQBIT_H_

#include "cplx.h"
#include "oneQbit.h"

struct FourByFour
{
  complex Matrix[5][5];
 
  FourByFour (complex a11, complex a12, complex a13, complex a14,
	       complex a21, complex a22, complex a23, complex a24,
	       complex a31, complex a32, complex a33, complex a34,
	       complex a41, complex a42, complex a43, complex a44 )
    {
      Matrix[1][1] = a11; Matrix[1][2] = a12; 
      Matrix[1][3] = a13; Matrix[1][4] = a14;
      Matrix[2][1] = a21; Matrix[2][2] = a22; 
      Matrix[2][3] = a23; Matrix[2][4] = a24;
      Matrix[3][1] = a31; Matrix[3][2] = a32; 
      Matrix[3][3] = a33; Matrix[3][4] = a34;
      Matrix[4][1] = a41; Matrix[4][2] = a42; 
      Matrix[4][3] = a43; Matrix[4][4] = a44;
    }
  FourByFour ()
  {
    int i; int j;
    for (i=1;i<=4;++i)
      for (j=1;j<=4;++j)
	Matrix[i][j]=complex(0.0,0.0);
    for(i=1;i<=4;++i)
      Matrix[i][i]=complex(1.0,0.0);
  }
};
  
FourByFour operator* (const FourByFour& A, const FourByFour& B);
FourByFour operator+ (const FourByFour& A, const FourByFour& B);
FourByFour operator* (complex z, FourByFour A);
complex det (const FourByFour& V);
FourByFour exp(const FourByFour& A);
FourByFour adj (const FourByFour& V);
FourByFour transpose (const FourByFour& V);
FourByFour inverse (FourByFour V);
FourByFour Tensor (OneQubitMap A, OneQubitMap B);

FourByFour operator* (const FourByFour& A, const FourByFour& B)
{
  FourByFour Answer;
  int i,j,k;
  //This is the product of two 4x4 matrices.
    for (i=1;i<=4;++i)
      for (j=1;j<=4;++j)
	{
	  Answer.Matrix[i][j]=0;
	  for (k=1; k<=4; ++k)
	    Answer.Matrix[i][j]+= 
	      A.Matrix[i][k]*B.Matrix[k][j];
	}
    return Answer;
}

FourByFour operator+ (const FourByFour& A, const FourByFour& B)
{
  FourByFour Answer;
  int i,j;

  for(i=1;i<=4;++i)
    for(j=1;j<=4;++j)
      Answer.Matrix[i][j]=A.Matrix[i][j]+B.Matrix[i][j];
  return Answer;
}

FourByFour operator* (complex z, FourByFour B)
{
  FourByFour Answer;
  int i,j;
    for (i=1;i<=4;++i)
      for (j=1;j<=4;++j)
	{
	  Answer.Matrix[i][j]=z*B.Matrix[i][j];
	}
    return Answer;
}

FourByFour exp (const FourByFour& A)
{
  int i;
  FourByFour Answer; //initializes to id.
  FourByFour Holder; //ditto
  for(i=1;i<=70;++i)
    {
      Holder=(1.0/i)*A*Holder;
      Answer=Answer+Holder;
    }
  return Answer;
}

complex det (const FourByFour& V)
{
  return (
	  V.Matrix[1][1]*V.Matrix[2][2]*V.Matrix[3][3]*V.Matrix[4][4]
	  -V.Matrix[1][1]*V.Matrix[2][2]*V.Matrix[3][4]*V.Matrix[4][3]
	  -V.Matrix[1][1]*V.Matrix[3][2]*V.Matrix[2][3]*V.Matrix[4][4]
	  +V.Matrix[1][1]*V.Matrix[3][2]*V.Matrix[2][4]*V.Matrix[4][3]
	  +V.Matrix[1][1]*V.Matrix[4][2]*V.Matrix[2][3]*V.Matrix[3][4]
	  -V.Matrix[1][1]*V.Matrix[4][2]*V.Matrix[2][4]*V.Matrix[3][3]
       	  -V.Matrix[2][1]*V.Matrix[1][2]*V.Matrix[3][3]*V.Matrix[4][4]
	  +V.Matrix[2][1]*V.Matrix[1][2]*V.Matrix[3][4]*V.Matrix[4][3]
	  +V.Matrix[2][1]*V.Matrix[3][2]*V.Matrix[1][3]*V.Matrix[4][4]
	  -V.Matrix[2][1]*V.Matrix[3][2]*V.Matrix[1][4]*V.Matrix[4][3]
	  -V.Matrix[2][1]*V.Matrix[4][2]*V.Matrix[1][3]*V.Matrix[3][4]
	  +V.Matrix[2][1]*V.Matrix[4][2]*V.Matrix[1][4]*V.Matrix[3][3]
	  +V.Matrix[3][1]*V.Matrix[1][2]*V.Matrix[2][3]*V.Matrix[4][4]
	  -V.Matrix[3][1]*V.Matrix[1][2]*V.Matrix[2][4]*V.Matrix[4][3]
	  -V.Matrix[3][1]*V.Matrix[2][2]*V.Matrix[1][3]*V.Matrix[4][4]
	  +V.Matrix[3][1]*V.Matrix[2][2]*V.Matrix[1][4]*V.Matrix[4][3]
	  +V.Matrix[3][1]*V.Matrix[4][2]*V.Matrix[1][3]*V.Matrix[2][4]
	  -V.Matrix[3][1]*V.Matrix[4][2]*V.Matrix[1][4]*V.Matrix[2][3]
	  -V.Matrix[4][1]*V.Matrix[1][2]*V.Matrix[2][3]*V.Matrix[3][4]
	  +V.Matrix[4][1]*V.Matrix[1][2]*V.Matrix[2][4]*V.Matrix[3][3]
	  +V.Matrix[4][1]*V.Matrix[2][2]*V.Matrix[1][3]*V.Matrix[3][4]
	  -V.Matrix[4][1]*V.Matrix[2][2]*V.Matrix[1][4]*V.Matrix[3][3]
	  -V.Matrix[4][1]*V.Matrix[3][2]*V.Matrix[1][3]*V.Matrix[2][4]
	  +V.Matrix[4][1]*V.Matrix[3][2]*V.Matrix[1][4]*V.Matrix[2][3]);
}

FourByFour transpose (const FourByFour& V)
{
  FourByFour U;
  int i,j;
  for(i=1;i<=4;++i)
    for(j=1;j<=4;++j)
      U.Matrix[i][j]=V.Matrix[j][i];
  return U;
}

FourByFour adj (const FourByFour& V)
{
  FourByFour U;
  int i,j;
  for(i=1;i<=4;++i)
    for(j=1;j<=4;++j)
      U.Matrix[i][j]=Bar(V.Matrix[j][i]);
  return U;
}

FourByFour inverse (FourByFour V)
{
  int i,j,k,l;
  int s,t;
  complex z,w;
  FourByFour U;  //U is constructed automatically as identity.

  //  for(s=1;s<=4;++s)
  //{ for(t=1;t<=4;++t)
  //  {
  //z=V.Matrix[s][t];
  //cout << z << " & ";
  //  }
  //cout << endl;
  //}  
  //cout << endl << endl;

  int ColumnNumber=0,RowNumber=0;  //Row & Col of last pivot entry.

  double EntrySize; double Largest; int m;
  
  for(j=1;j<=4;++j)
    {
      EntrySize=0.0;
      Largest=0.0;
      m=0;
      for(i=RowNumber+1;i<=4;++i)
	{
	  EntrySize=fabs(Re(V.Matrix[i][j]))+fabs(Im(V.Matrix[i][j]));
	  if(EntrySize>Largest)
	    {m=i; Largest=EntrySize;}
	} // Matrix[m][j] now holds largest eligible entry of col j
      if(Largest < 0.005)  //Clear column j if no pivot.
	{
	  for(i=RowNumber+1;i<=4;++i)
	    V.Matrix[i][j]=complex(0.0,0.0);
	  for(i=1;i<=RowNumber;++i)
	    {   
	      if(fabs(Re(V.Matrix[i][j])) + fabs(Im(V.Matrix[i][j])) < 0.005)
		V.Matrix[i][j]=complex(0.0,0.0);
	    }
	}
      if(Largest >= 0.005)  //Matrix[i][j] is declared a pivot.
	{ //flip row of pivot to highest left possible so as to not interfere.
	  RowNumber++; ColumnNumber=j;
	  z=V.Matrix[m][ColumnNumber];
	  // cout << "Pivot at (" << m << "," << RowNumber
	  //   << ")" << " of " << z << endl;
	  //  {for(s=1;s<=4;++s)
	  //{ for(t=1;t<=4;++t)
	  //  {
	  //z=V.Matrix[s][t];
	  //cout << z << " & ";
	  //  }
	  //cout << endl;
	  //}}
	  // cout << endl;
	  for(k=1;k<=4;++k)
	    {
	      z=V.Matrix[RowNumber][k];
	      V.Matrix[RowNumber][k]=V.Matrix[m][k];
	      V.Matrix[m][k]=z;
	      w=U.Matrix[RowNumber][k];
	      U.Matrix[RowNumber][k]=U.Matrix[m][k];
	      U.Matrix[m][k]=w;
	    }
	  // cout << "After the Row Swap we get this: " << endl;
	  //{for(s=1;s<=4;++s)
	  //{ for(t=1;t<=4;++t)
	  //  {
	  //z=V.Matrix[s][t];
	  //cout << z << " & ";
	  //  }
	  //cout << endl;
	  //}}
	  // cout << endl;
	  //  cout << "j is now " << j << endl;
	  //Clear row by dividing out pivot.
	  z=V.Matrix[RowNumber][ColumnNumber];
	  for(l=1;l<=4;++l)
	    {
	      V.Matrix[RowNumber][l]=V.Matrix[RowNumber][l]/z;
	      U.Matrix[RowNumber][l]=U.Matrix[RowNumber][l]/z;
	    }
	  // cout << "After clearing out the pivot row, this results:";
	  //  {for(s=1;s<=4;++s)
	  //{ for(t=1;t<=4;++t)
	  //  {
	  //z=V.Matrix[s][t];
	  //cout << z << " & ";
	  //  }
	  //cout << endl;
	  //}}

	  for(k=1;k<=4;++k)  //Use pivot to clear its column.
	    {
	      if(k!=RowNumber)
	      {
		z=V.Matrix[k][ColumnNumber];
		for(l=1;l<=4;++l)
		  {
		    V.Matrix[k][l]=V.Matrix[k][l]-z*V.Matrix[RowNumber][l];
		    U.Matrix[k][l]=U.Matrix[k][l]-z*U.Matrix[RowNumber][l];
		  }
	      }
	    }
	}
      //   cout << "Before Loop EndS: j is now still " << j << endl;
    }
  return U;
}


FourByFour Tensor (OneQubitMap A, OneQubitMap B)
{
  FourByFour Answer;

  //This is the tensor product of A and B.
  Answer.Matrix[1][1]=TopLeft(A) * TopLeft(B);
  Answer.Matrix[1][2]=TopLeft(A) * TopRight(B);
  Answer.Matrix[2][1]=TopLeft(A) * BotLeft(B);
  Answer.Matrix[2][2]=TopLeft(A) * BotRight(B);
  Answer.Matrix[1][3]=TopRight(A) * TopLeft(B);
  Answer.Matrix[1][4]=TopRight(A) * TopRight(B);
  Answer.Matrix[2][3]=TopRight(A) * BotLeft(B);
  Answer.Matrix[2][4]=TopRight(A) * BotRight(B);
  Answer.Matrix[3][1]=BotLeft(A) * TopLeft(B);
  Answer.Matrix[3][2]=BotLeft(A) * TopRight(B);
  Answer.Matrix[4][1]=BotLeft(A) * BotLeft(B);
  Answer.Matrix[4][2]=BotLeft(A) * BotRight(B);
  Answer.Matrix[3][3]=BotRight(A) * TopLeft(B);
  Answer.Matrix[3][4]=BotRight(A) * TopRight(B);
  Answer.Matrix[4][3]=BotRight(A) * BotLeft(B);
  Answer.Matrix[4][4]=BotRight(A) * BotRight(B);

  return Answer;
};

ostream& operator<<(ostream& out, FourByFour& V)
{ 
  int i,j,k;
  complex z;
  ostream& answer=out;
    for (i=1;i<=4;++i)
    {    
      for (j=1;j<=4;++j)
	{
	  z = V.Matrix[i][j];
	  answer << z << "   ";
	}
      answer << endl;
    };
    return answer;
}

struct TensorDecomp
{  OneQubitMap FirstFactor; OneQubitMap SecondFactor; 
  TensorDecomp (const FourByFour& U);
};

//  Code must handle case block is 0 correctly.

TensorDecomp:: TensorDecomp (const FourByFour& U)
{
  const OneQubitMap OneQubitIdentity; //constructor defaults to Id.

  double UpLeftTraceNorm;
  complex a11, a12, a21, a22;
  OneQubitMap A;
  OneQubitMap B;
  OneQubitMap C;
  complex z;

  UpLeftTraceNorm = norm(U.Matrix[1][1]) + norm(U.Matrix[1][2])
    + norm(U.Matrix[2][1]) + norm(U.Matrix[2][2]);
  
  B = OneQubitMap (U.Matrix[1][1], U.Matrix[1][2], U.Matrix[2][1], 
		   U.Matrix[2][2]);
  //cout << "The UpLeftTraceNorm is " << UpLeftTraceNorm <<"." <<  endl;
  if(UpLeftTraceNorm < 0.90)
    {
      B = OneQubitMap (U.Matrix[1][3], U.Matrix[1][4], 
		       U.Matrix[2][3], U.Matrix[2][4]);
    }
  //B should now be correct, up to constant real multiple of col norm.
  z=norm(TopLeft(B))+norm(BotLeft(B));
  C=sqrt((1/Re(z)))*B;
  B=C;
  //Fixed that.
  SecondFactor = B;
  FourByFour V=U;
  V = U* Tensor( OneQubitIdentity, inverse(B));
  //cout << "The matrix V is as follows:" << endl;
  //cout << V;
  a11 = (0.50)*(V.Matrix[1][1]+V.Matrix[2][2]);
  a12 = (0.50)*(V.Matrix[1][3]+V.Matrix[2][4]);
  a21 = (0.50)*(V.Matrix[3][1]+V.Matrix[4][2]);
  a22 = (0.50)*(V.Matrix[3][3]+V.Matrix[4][4]);
  FirstFactor = OneQubitMap( a11, a12, a21, a22);
};

#endif

